#include <math.h>
#include <stdlib.h>
#include <time.h>

//////////////////////////////////////////////////////////////////////////////
// defines
//////////////////////////////////////////////////////////////////////////////
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif


#define BLK_SIZE  1024 
#define MAXLONGINT 32768 
#define NUM_TRIES 10
#define ZAEHLER
//#define SHOWMBR

#include "blk_file.h"
#include "linlist.h"
#include "rtree.h"

class DATA
{
    int dimension;
public:
    float *data;                       // Vector
    float distanz;                     // anything
    float *get_mbr();                  // returns MBR of the object
    float get_area()                   // returns the area of the MBR
    { return 0.0; }                    // for vectors always 0.0
    void read_from_buffer(char *buffer);   
                                       // reads data from buffer
    void write_to_buffer(char *buffer);   
                                       // writes data to buffer
    int get_size();                    // returns amount of needed space
    void print();

    virtual DATA & operator = (DATA &_d);
    virtual bool operator < (DATA &_d);
    virtual bool operator > (DATA &_d);

    DATA(int *dimension = &RTDataNode__dimension);
    virtual ~DATA();
};


void error(char *t, bool ex)
{
    fprintf(stderr, t);
    if (ex)
	exit(0);
}

//////////////////////////////////////////////////////////////////////////////
// DATA
//////////////////////////////////////////////////////////////////////////////

DATA::DATA(int *_dimension)
{
    dimension = *((int *)_dimension);
    data = new float[dimension];
}

DATA::~DATA()
{
    delete [] data;
}

DATA & DATA::operator = (DATA &_d)
{
    dimension = _d.dimension;
    distanz = _d.distanz;
    memcpy(data, _d.data, sizeof(float) * dimension);

    return *this;
}

bool DATA::operator < (DATA &_d)
{
    return (distanz < _d.distanz);
}

bool DATA::operator > (DATA &_d)
{
    return (distanz > _d.distanz);
}

void DATA::read_from_buffer(char *buffer)
{
    int i;

    i = dimension*sizeof(float);
    memcpy(data, buffer, i);
    memcpy(&distanz, &buffer[i], sizeof(int));
}

void DATA::write_to_buffer(char *buffer)
{
    int i;

    i = dimension*sizeof(float);
    memcpy(buffer, data, i);
    memcpy(&buffer[i], &distanz, sizeof(int));
}

int DATA::get_size()
{
    return dimension * sizeof(float) + sizeof(int);
}

void DATA::print()
{
 int i;

 printf("( ");

 for(i = 0; i < dimension; i++)
   printf("%7.5f ",data[i]);

 printf(")\n");

}

float *DATA::get_mbr()
// fuer Punktdaten trivial: untere_grenze == obere_grenze
{
    int i;
    float *f;

    f = new float[2*dimension];
    for (i = 0; i < dimension; i++)
	f[2*i] = f[2*i+1] = data[i];

    return f;
}


////////////////////////////////////////////////////////////////
// main()
///////////////////////////////////////////////////////////////

void pm(float *mbr)
{
    int i;

    for (i = 0; i < 8; i++)
	printf("%f..%f ", mbr[2*i], mbr[2*i+1]);
    printf("\n");
}

float page_access;

#ifdef SHOWMBR
	int split_000 = 0;
#endif


class Timer
{
    float dist_time_cpu; 
    float dist_time; 

    clock_t clock_start, clock_end;
    time_t time_start, time_end;
public:
    void start();
    void end();
};

void Timer::start()
{
    clock_start = clock();
    time (&time_start);
}


void Timer::end()
{
    time (&time_end);
    clock_end = clock ();
    dist_time_cpu = (float) (clock_end - clock_start) / 
	(float) CLOCKS_PER_SEC;
    dist_time = difftime(time_end, time_start);
    
    printf("%f sec cpu,   %f sec real\n", 
	   dist_time_cpu, dist_time);
}




main(int argc, char *argv[])
{
    RTree<DATA> *rt, *rtneu;
    DATA *d, *d1, *p;
     float i, j, *r,i0,i1,i2,i3,i4;
    int k, l, dim, z, n, nn,zahl, u, v, w;
    
    char filename[30]; 
    char dummy;
    SortedLinList<DATA> *res;
    Timer t;

    FILE *f,*f1;


#ifdef XXX
    float *fa, *fb;
    float dist_time_cpu; 
    float dist_time; 

    clock_t clock_start, clock_end;
    time_t time_start, time_end;

    fa = new float[1001000];
    fb = new float[1001000];

    clock_start = clock();
    time (&time_start);



    for (k = 0; k < 1000; k++)
    {
	for (l = 0; l < 200; l++)
	    section(10, &(fa[1000*k+l]), &(fb[1000*k+l]));
	if (k % 100 == 0)
	    printf("%d\n", k);
    }

    BlockFile *bf;
    char *c;

    bf = new BlockFile("test.rt", 8192);
    n = bf->get_num_of_blocks();
    c = new char[10000];
    for (z = 0; z < 2; z++)
    {
	for (k = 0; k < 1000; k++)
	{
	    for (l = 0; l < n; l++)
		bf->read_block(c, l);
	}
    }
    
    z--;
    time (&time_end);
    clock_end = clock ();
    dist_time_cpu = (float) (clock_end - clock_start) / 
	(float) CLOCKS_PER_SEC;
    dist_time = difftime(time_end, time_start);

    printf("%f sec cpu / Block,   %f sec real / Block\n", dist_time_cpu / (float)(n*z), dist_time / (float)(n*z));

    exit(0);

#endif   // XXX


    dim = atoi(argv[2]);

    if (dim < 2 || dim > 1000)
	error("dimension out of range", TRUE);

    if (!strcmp(argv[1], "c"))
    {
        // neuen Baum erzeugen
        rt = new RTree<DATA>("test.rt",BLK_SIZE, 4, dim);
    }
    else if (!strcmp(argv[1], "fe"))
    {
        // alten Baum laden
        rt = new RTree<DATA>("test.rt", 4);
        // Baum fuellen
	i = 46657.0;
	j = pow(i, (1.0/dim))+0.5;
	k = (int) j;                          // Basis 
	l = (int) pow((float)k, (float)dim);  // Grenze


	for (n = 0; n < l; n++)
	{
	    d = new DATA(&dim);
	    
	    w = n;
	    for (z = dim-1; z >=0; z--)
	    {
		u = (int) pow((float)k, (float)z);
		v = w / u;
		d->data[z] = ((float) v) / (float) (k-1);
		w -= v*u;
	    }

	    if (n % 50 == 0)
		printf("did %d\n", n);

	    rt->insert(d);
	    delete d;
	}
    }
    else if (!strcmp(argv[1], "ferand"))
    {
        // alten Baum laden
        rt = new RTree<DATA>("test.rt.fe", 4);
	rtneu = new RTree<DATA>("test.rt",4);

        // Baum fuellen
        for(i0 = 0.0; i0 < 100000; i0++)
	{
		k = (int)  (((float) rand() / MAXLONGINT) * 100000);
		d = rt->get(k);
		rtneu->insert(d);
		delete d;

		printf("did %3f\n", i0);
	}
			
    }    	


    else if (!strcmp(argv[1], "cl"))
    {
	srand(456789);	

        // alten Baum laden
        rt = new RTree<DATA>("test.rt", 4);
        
	// Baum fuellen
	// 1.000 Cluster-Punkte
	// jeweils 100 Punkte in einer eps - Umgebungs
	// eps = 1.0e-3

	for(i0 = 0.0; i0 < 400; i0+= 1.0 )
	{
                // neuen Punkt generieren 
		d = new DATA(&dim);
		for(l = 0; l < dim; l++)
           		d->data[l] = (float) rand() / MAXLONGINT;

            	for(i1 = 0.0; i1 < 25.0; i1 += 1.0)
		// um diesen Punkt in der eps-Umgebung 
		// 100 Punkte setzen (zuf"allig setzen
		{
			d1 = new DATA(&dim);
			for(l = 0; l < dim; l++)
				d1->data[l] = d->data[l] + 
				 (1.0e-2 - (5.0e-3 * (float) rand() / MAXLONGINT));
			printf("%f:", i0 *100+i1);
			d1->print();
   	       		rt->insert(d1);
                	delete d1;
		}
		d->print();
		rt->insert(d);	
		delete d;    
		printf("did %3f\n", i0);
	 }
    }     
    else if (!strcmp(argv[1], "fd"))
    {
        // alten Baum laden
        rt = new RTree<DATA>("test.rt", 4);
	
        srand(100366);
        // Baum fuellen
        for (j = 0.0; j < 100000; j++ )
        {
	    d = new DATA(&dim);
	    d->data[0] = (float) rand() / 1.0e10;
	    for(l = 1; l < dim; l++)
		d->data[l] = d->data[0] + 
		    (0.001 * (float) rand() / 1.0e10  - 0.0005);
	    
            rt->insert(d);
            delete d;
	    
	    if ((int) j % 50 == 0)
	    	printf("did %5f\n",j);
	}
    }
    else if (!strcmp(argv[1], "re"))
    {
	int anz;
	FILE *fp;
	int wie_oft;
	float ff[100];
	
	anz = atoi(argv[3]);
	
	// alten Baum laden
	rt = new RTree<DATA>("test.rt", 4);
	
	// Baum fuellen
	
	wie_oft = 16 / dim;
	if (wie_oft == 0)
	    error("cannot read vectors larger than dim 16 from founew.dump",
		  TRUE);
	
	fp = fopen("founew.dump", "rb");
	
		printf("reading %d vectors into index of dimension %d\n",
		       anz, dim);
	
	if (fp == NULL)
	    error("error opening founew.dump", TRUE);
	t.start();
	for (j = 0.0; j < anz; j++ )
	{
	    d = new DATA(&dim);
	    fread(ff, sizeof(float), dim, fp);
	    for(l = 0; l < dim; l++)
	    {
		d->data[l] = ff[l];
		d->data[l] += ff[l] * ((float) rand() / 100000000.0);
	    }
	    
	    rt->insert(d);
	    
	    if ((int) j % 50 == 0)
		printf("did %5f\n",j);
	    delete d;
	}
	t.end();
	fclose(fp);
    }
    else if (!strcmp(argv[1], "fr"))
    {
	int anz;
	
	anz = atoi(argv[3]);

	// alten Baum laden
	rt = new RTree<DATA>("test.rt", 4);
	srand(2314);
	// Baum fuellen
	
	t.start();
        for (j = 0.0; j < anz; j++ )
        {
	    d = new DATA(&dim);
	    for(l = 0; l < dim; l++)
		d->data[l] = ((float) rand() / (float)MAXLONGINT);
            	            
	    rt->insert(d);
	    delete d;

	    if ((int) j % 50 == 0)
		printf("did %5f\n",j);
        }
	t.end();


    }
    else if (!strcmp(argv[1], "rq"))
    {
        // alten Baum laden
        rt = new RTree<DATA>("test.rt", 4);

	printf("\n\ntree is filled now --> range query... (0.04-0.06)^5 \n");

	r = new float[2 * dim];

	for(l = 0; l < dim; l++)
	{
		r[2*l] = 0.20;
		r[2*l+1] = 0.20000000001;
	}

	rt->region(r);
        delete r;
    }
    else if (!strcmp(argv[1], "nrq"))
    {
	int num_quer;
	float left_point, query_range;

        // alten Baum laden
 	page_access = 0;
	r = new float[2 * dim];
	num_quer = 10;
	query_range = pow(0.001, 1.0/((float)dim));

	printf("%d range queries: Kantenlaenge %f...\n", 
	       num_quer, query_range);
	
	for (k = 0; k < num_quer; k++)
	{
	    rt = new RTree<DATA>("test.rt", 4);
	    for(l = 0; l < dim; l++)
	    {
		i = ((float) rand() / (float)MAXLONGINT);
		left_point = (1.0 - query_range) * i;
		r[2*l] = left_point;
		r[2*l+1] = left_point + query_range;
	    }

	    rt->region(r);
	    delete rt;
	}
        delete r;

	printf("Anzahl der Plattenzugriffe: %f \n",
	       page_access/(float)num_quer);
    }
    else if (!strcmp(argv[1], "rqnn"))
    {
        // alten Baum laden
        rt = new RTree<DATA>("test.rt", 4);
	filename = "dimvgl";
	strcat(filename,argv[2]);

	f = fopen(filename,"w");
	printf("\n\ntree is filled now --> range query... (0.04-0.06)^5 \n");
	for(nn = 0; nn < 30; nn++)
	{
	
	page_access = 0.0;
	res = new SortedLinList<DATA>; 
	res->set_sorting(TRUE);	
	
	page_access = 0.0;
	r = new float[2 * dim];

	for(l = 0; l < dim; l++)
	{
		r[2*l] = (float) rand() / MAXLONGINT;
		r[2*l+1] = r[2*l] + 0.000000000001;
	}
        
	rt->rangeQuery(r,res);
        //delete r;
         
   	printf("RQ:page_access: %f Treffer-Zahl %d \n",
	       page_access,res->get_num());
        

	fprintf(f,"%d;%f; ",res->get_num(),page_access);	


	d = new DATA(&dim);
	p = new DATA(&dim);
	for(l = 0; l < dim; l++)
	{
		d->data[l] = (r[2*l]+r[2*l+1]) / 2 ;
		p->data[l] = 1.0e10;
	}

	page_access = 0.0;

	rt->NearestNeighborQuery(d,p);	 
	printf("NN:page_access : %f ",page_access);
	p->print();

	fprintf(f,"%f ",page_access);
	//d->print();
	//p->print();
	for(l = 0; l < dim; l++)
	{
		p->data[l] = 1.0e10;
	}
	delete res;
	page_access = 0.0;
	res = new SortedLinList<DATA>; 
	res->set_sorting(TRUE);
	
	rt->range_nnQuery(r,res,p);        
	printf("RQ+NN:page_access: %f ",page_access);        
	p->print();
        fprintf(f,"; %f \n",page_access);
	printf("\n\n");

	delete p;
	delete d;
	delete r;
	delete res;
	}
	
	rt->printinfo(f);
	fclose(f);	
    }
    else if (!strcmp(argv[1], "overlap"))
    {
	int i, j, all, num;
	int nodes_b[20];   // Anzahl der Knoten pro Tiefe
	int nodes_t[20];   
	int nodes_s[20];   
	int nodes_a[20];   
	float *mbr;
	float mi[30], ma[30], ex[30];
	DATA *d;

	rt = new RTree<DATA>("test.rt", 4);
	
	for (i = 0; i < 20; i++)
	    nodes_t[i] = nodes_s[i] = nodes_b[i] = nodes_a[i] = 0;

	// bestimme Anzahl der Knoten pro level
	rt->nodes(nodes_a);

	// bestimme 1 % Anfragepunkte
	all = rt->get_num();
	num = all / 100;

	printf("querying %d of %d data\n", num, all);
	rt->load_root();
	mbr = rt->root_ptr->get_mbr();

	for (i = 0; i < dim; i++)
	{
	    ma[i] = -MAXREAL;
	    mi[i] = MAXREAL;
	}

	for (i = 0; i < dim; i++)
	{
	    ma[i] = max(mbr[2*i+1], ma[i]);
	    mi[i] = min(mbr[2*i], mi[i]);
	}

	for (i = 0; i < dim; i++)
	    ex[i] = ma[i] - mi[i];

	for (i = 0; i < num; i++)
	{
//        d = rt->get(rand()/10);
	    d = new DATA(&dim);
	    for(l = 0; l < dim; l++)
		d->data[l] = ma[l] - ex[l] * ((float) rand() / (float)MAXLONGINT);
	   
	    for (j = 0; j < 20; j++)
		nodes_b[j] = 0;      

	    rt->overlapping(d->data, nodes_b);

	    for (j = 0; j < 20; j++)
	    {
		nodes_s[j] += nodes_b[j];
		if (nodes_b[j] > 1)
		    nodes_t[j]++;
	    }
	    delete d;
	}

	for (i = 0; i < 5; i++)
	{
	      printf("nodes accessed on level %d: %f,  %f of %d\n", i, 
		 (float)nodes_t[i]/(float)num, 
		 (float)nodes_s[i]/(float)num, nodes_a[i]);
	}
    }
    else if (!strcmp(argv[1], "w"))
    {
	rt = new RTree<DATA>("test.rt", 4);
	f = fopen("mbr","w");
	rt->writeinfo(f);
	fclose(f);
    }
    else if (!strcmp(argv[1], "nn"))
    {
	// alten Baum laden
	srand(12345);
	rt = new RTree<DATA>("test.rt", 4);
	
	printf("tree is filled now --> nearest Narbor query von (0.05)^5 \n");
        for(z = 0; z < 1; z++)
	{
	    page_access = 0.0;
	    d = new DATA(&dim);
	    for(l = 0; l < dim; l++)
		d->data[l] = ((float) rand() / (float)MAXLONGINT);
	    
	    printf("Query-Point :");
	    d->print();
	    
	    p = new DATA(&dim);
	    for(l = 0; l < dim; l++)  
		p->data[ l ] = 2.0;
	    
	    rt->NearestNeighborQuery( d, p);

	    printf("\nNearest:");
	    p->print(); 
	    delete p;
	    delete d;
	    
	    printf("\nQuery-Nummer: %d Plattenzugriffe: %f \n\n",
		   z, page_access);	
	}
    }
    else if (!strcmp(argv[1], "gknn"))
    {
	float all_page;
	int cache_size;

        // alten Baum laden
	srand(140267);

	k = atoi(argv[3]);
	rt = new RTree<DATA>("test.rt", 4);

	// berechne Gewichte fuer Cache
	cache_size = (int) ((12 * dim * 
			    (rt->num_of_dnodes+rt->num_of_inodes)) / BLK_SIZE);
	rt->set_node_weight(cache_size);

	t.start();
	all_page = 0.0;
	for(nn = 0; nn < 10; nn++)
	{
	    page_access = 0.0;

 	    d = rt->get(rand()/10);
	    d->distanz = MAXREAL;

//	    printf("Query-Point :");
//	    d->print();
 
	    res = new SortedLinList<DATA>;
	    rt->k_NearestNeighborQuery(d, k, res);

//	    printf("\n%d Nearest:\n",k);
//	    for(n = 0; n < k; n++)
//	    {
//        	printf("%1d.Nearest: Distanz: %f Point: ",n,
//		       res->get(n)->distanz);
//		res->get(n)->print();
//	    } 	    
//	    printf("page_accesses: %f \n\n",page_access);

	    delete d;
	    delete res;

	    all_page += page_access;
	    printf(".");
	    fflush(stdout);
	}
	printf("average (10) page_accesses (%d NN): %f \n\n", 
	       k, all_page / 10.0);
	t.end();
    }
    else if (!strcmp(argv[1], "pq"))
    {
	float all_page;
	int l;

        // alten Baum laden
	srand(2314);

	rt = new RTree<DATA>("test.rt", 4);

	t.start();
	all_page = 0.0;
	for (nn = 0; nn < NUM_TRIES; nn++)
	{
	    page_access = 0.0;

	    l = rand()/10;
	    printf("/%d/\n", l);
 	    d = rt->get(l);

//	    printf("Query-Point :");
//          d->print();

	    rt->point_query(d->data);

	    delete d;

	    all_page += page_access;
	    printf(".");
	    fflush(stdout);
	}
	printf("average (%d) page_accesses: %f \n\n", 
	       NUM_TRIES, all_page / NUM_TRIES);
	t.end();
    }
    else if (!strcmp(argv[1], "knn"))
    {
	float all_page;
	int cache_size;

        // alten Baum laden 
	srand(907932);

	k = atoi(argv[3]);
	rt = new RTree<DATA>("test.rt", 4);

	// berechne Gewichte fuer Cache
	cache_size = (int) ((12 * dim * 
			    (rt->num_of_dnodes+rt->num_of_inodes)) / BLK_SIZE);
	rt->set_node_weight(cache_size);

	t.start();
	all_page = 0.0;
	for(nn = 0; nn < 10; nn++)
	{
	    page_access = 0.0;

	    d = new DATA(&dim);
	    for(l = 0; l < dim; l++)
		d->data[l] = ((float) rand() / (float)MAXLONGINT);
	    d->distanz = MAXREAL;

//	    printf("Query-Point :");
//	    d->print();
 
	    res = new SortedLinList<DATA>;
	    rt->k_NearestNeighborQuery(d, k, res);

// 	    printf("\n%d Nearest:\n",k);
// 	    for(n = 0; n < k; n++)
// 	    {
//         	printf("%1d.Nearest: Distanz: %f Point: ",n,
// 		       res->get(n)->distanz);
// 		res->get(n)->print();
// 	    } 
//	    
//	    printf("page_accesses: %f \n\n",page_access);

	    delete d;
	    delete res;

	    all_page += page_access;
	    printf(".");
	    fflush(stdout);
	}
	printf("average (10) page_accesses (%d NN): %f \n\n", 
	       k, all_page / 10.0);
	t.end();
    }
    else if (!strcmp(argv[1], "n1"))
    {
        f=fopen("output.1","w");
	fprintf(f,"Test nearest Nabour Query: 5-dim, 100000 Datas, zufallsverteilung\n");
	
	//printf("tree is filled now --> nearest Narbor query von (0.05)^5 \n");
        for(z = 0; z < 100; z++)
	{
	    rt = new RTree<DATA>("test.rt", 4); 
	    page_access = 0.0;
	    d = new DATA(&dim);
	    for(l = 0; l < dim; l++)
		d->data[ l ] = (float) rand() / 1.0e10;   
	    
	    
	    printf("Query-Point :");
	    d->print();
	    
	    p = new DATA(&dim);
	    for(l = 0; l < dim; l++)  
		p->data[ l ] = 2.0;
	    
	    
	    rt->NearestNeighborQuery( d, p);
	    
	    
	    printf("nNearest:");
	    p->print(); 
	    delete p;
	    delete d;
	    
	    printf("\nQuery-Nummer: %d Plattenzugriffe: %f \n\n",z,page_access);
	    fprintf(f,"%d;%f\n",z,page_access);
	    delete rt;
	}
	fclose(f);
    }
    else if (!strcmp(argv[1], "so"))
    {
        // alten Baum laden
        rt = new RTree<DATA>("test.rt", 4);
	
	for (k = 0; k < rt->get_num(); k++)
	{
	    d = rt->get(k);
	    if ((d != NULL) && (k % 100 == 0))
	    {
		printf("%d:", k);
		d->print();
		printf("\n");
		delete d;
            }
        }
    }
    else if (!strcmp(argv[1], "so1"))
    {
        // alten Baum laden
        rt = new RTree<DATA>("test.rt", 4);
        f = fopen("out.out","w");
	for (k = 0; k < 15000; k++)
	{
	    d = rt->get(k);
	    if ((d != NULL) )
	    {
		printf("%d:", k);
		d->print();
		printf("\n");
		fprintf(f,"%f ; %f \n",d->data[0],d->data[1]);
		delete d;

            }
        }
	fclose(f);
    }
              
    delete rt;

    return 0;
}
